#define _USE_MATH_DEFINES
#include <cmath>
#include <iostream>
#include <filesystem>
#include <set>

#define TINYOBJLOADER_IMPLEMENTATION
#include "tiny_obj_loader.h"

using namespace std;
namespace fs = std::filesystem;

void loadOBJ(const string &meshName, vector<float> &posBuf, vector<int> &indBuf)
{
    // Load geometry
    tinyobj::attrib_t attrib;
    std::vector<tinyobj::shape_t> shapes;
    std::vector<tinyobj::material_t> materials;
    string warnStr, errStr;
    bool rc = tinyobj::LoadObj(&attrib, &shapes, &materials, &warnStr, &errStr, meshName.c_str());
    if(!rc) {
        cerr << errStr << endl;
    } else {
        // Get positions
        const vector<float>& tmpPos = attrib.GetVertices();
        posBuf.insert(posBuf.end(), tmpPos.begin(), tmpPos.end());
        // Get indices from all shapes
        for(int k = 0; k < shapes.size(); ++k) {
            const vector<tinyobj::index_t> &tmpInd = shapes[k].mesh.indices;
            for(int i = 0; i < tmpInd.size(); ++i) {
                indBuf.push_back(tmpInd[i].vertex_index);
            }
        }
    }
}

// For testing the import
void exportOBJ(const std::string& filename,
               const std::vector<float>& positions,
               const std::vector<int>& indices) {
    std::ofstream outFile(filename);
    if (!outFile) {
        std::cerr << "Error: Cannot open file " << filename << " for writing.\n";
        return;
    }
    
    // Write vertex positions
    for (size_t i = 0; i < positions.size(); i += 3) {
        outFile << "v " << positions[i] << " " << positions[i + 1] << " " << positions[i + 2] << "\n";
    }
    
    // Write face indices (OBJ format uses 1-based indexing)
    for (size_t i = 0; i < indices.size(); i += 3) {
        outFile << "f "
                << (indices[i + 0] + 1) << " "
                << (indices[i + 1] + 1) << " "
                << (indices[i + 2] + 1) << "\n";
    }
    
    outFile.close();
    //std::cout << "OBJ file successfully written to " << filename << "\n";
}

float dot(const float* a, const float* b) {
    return a[0] * b[0] + a[1] * b[1] + a[2] * b[2];
}

void subtract(const float* a, const float* b, float* result) {
    result[0] = a[0] - b[0];
    result[1] = a[1] - b[1];
    result[2] = a[2] - b[2];
}

void cross(const float* a, const float* b, float* result) {
    result[0] = a[1] * b[2] - a[2] * b[1];
    result[1] = a[2] * b[0] - a[0] * b[2];
    result[2] = a[0] * b[1] - a[1] * b[0];
}

float lengthSquared(const float* v) {
    return dot(v, v);
}

float pointTriangleDistance(const float* p, const float* a, const float* b, const float* c) {
    float ab[3], ac[3], ap[3], bc[3], bp[3], cp[3];
    subtract(b, a, ab);
    subtract(c, a, ac);
    subtract(p, a, ap);
    subtract(c, b, bc);
    subtract(p, b, bp);
    subtract(p, c, cp);
    
    float d1 = dot(ab, ap);
    float d2 = dot(ac, ap);
    float d3 = dot(ab, bp);
    float d4 = dot(ac, bp);
    float d5 = dot(ab, cp);
    float d6 = dot(ac, cp);
    
    if (d1 <= 0.0f && d2 <= 0.0f) return std::sqrt(lengthSquared(ap));
    if (d3 >= 0.0f && d4 <= d3)   return std::sqrt(lengthSquared(bp));
    if (d6 >= 0.0f && d5 <= d6)   return std::sqrt(lengthSquared(cp));
    
    float vc = d1 * d4 - d3 * d2;
    if (vc <= 0.0f && d1 >= 0.0f && d3 <= 0.0f)
        return std::sqrt(lengthSquared(ap) - d1 * d1 / lengthSquared(ab));
    
    float vb = d5 * d2 - d1 * d6;
    if (vb <= 0.0f && d2 >= 0.0f && d6 <= 0.0f)
        return std::sqrt(lengthSquared(ap) - d2 * d2 / lengthSquared(ac));
    
    float va = d3 * d6 - d5 * d4;
    if (va <= 0.0f && (d4 - d3) >= 0.0f && (d5 - d6) >= 0.0f)
        return std::sqrt(lengthSquared(bp) - (d4 - d3) * (d4 - d3) / lengthSquared(bc));
    
    float n[3];
    cross(ab, ac, n);
    float nLenSquared = lengthSquared(n);
    return std::fabs(dot(ap, n)) / std::sqrt(nLenSquared);
}

float computeMeshDistance(const std::vector<float>& positions, const std::vector<int>& indices, const float* point, double tol) {
    float minDist = std::numeric_limits<float>::max();
    for (size_t i = 0; i < indices.size(); i += 3) {
        const float* a = &positions[3 * indices[i + 0]];
        const float* b = &positions[3 * indices[i + 1]];
        const float* c = &positions[3 * indices[i + 2]];
        
        float centroid[3] = {
            (a[0] + b[0] + c[0]) / 3.0f,
            (a[1] + b[1] + c[1]) / 3.0f,
            (a[2] + b[2] + c[2]) / 3.0f
        };
        
        float radiusSquared = lengthSquared(a) * 2.0f; // Increased radius
        
        float pointToCentroid[3];
        subtract(point, centroid, pointToCentroid);
        if (lengthSquared(pointToCentroid) > radiusSquared) {
            continue; // Skip if point is outside the bounding sphere
        }
        
        float d = pointTriangleDistance(point, a, b, c);
        if (d < minDist) {
            minDist = d;
//            if(minDist < tol) {
//                return minDist; // early exit
//            }
        }
    }
    return minDist;
}

bool objscmp(const string &folder1, const string &folder2, double tol)
{
    // https://stackoverflow.com/questions/62409409/how-to-make-stdfilesystemdirectory-iterator-to-list-filenames-in-order
    set<fs::path> sortedFiles1;
    set<fs::path> sortedFiles2;
    cout << folder1 << endl;
    for(const auto &entry : fs::directory_iterator(folder1)) {
        if(entry.path().extension() == ".obj") {
            //cout << entry.path() << endl;
            sortedFiles1.insert(entry.path());
        }
    }
    cout << folder2 << endl;
    for(const auto &entry : fs::directory_iterator(folder2)) {
        if(entry.path().extension() == ".obj") {
            //cout << entry.path() << endl;
            sortedFiles2.insert(entry.path());
        }
    }
    if(sortedFiles1.size() != sortedFiles2.size()) {
        cout << "The directories do not have the same number of obj files." << endl;
        return false;
    }
    // Now the directories do have the same number of files
    
    // Run test
    bool match = true;
    int nobjs = (int)sortedFiles1.size();
    auto file1 = sortedFiles1.cbegin();
    auto file2 = sortedFiles2.cbegin();
    for(int k = 0; k < nobjs; ++k) {
        cout << file1->filename() << ": ";
        if(file1->filename() != file2->filename()) {
            cout << "Filenames do not match: " << file1->filename() << ", " << file2->filename() << endl;
        }
        //cout << entry1->filename() << endl;
        vector<float> posBuf1;
        vector<float> posBuf2;
        vector<int> indBuf1;
        vector<int> indBuf2;
        loadOBJ(file1->string(), posBuf1, indBuf1);
        loadOBJ(file2->string(), posBuf2, indBuf2);
        
        // Check if imported successfully
        //exportOBJ("foo1.obj", posBuf1, indBuf1);
        //exportOBJ("foo2.obj", posBuf2, indBuf2);
        
        if(posBuf1.size() != posBuf2.size()) {
            cout << file1->filename() << " and " << file2->filename() << " do not have the same number of vertices." << endl;
            return false;
        }
        
        // Check mesh1 verts against mesh2
        float distMax = 0;
        for(size_t i = 0; i < posBuf1.size(); i += 3) {
            float dist = computeMeshDistance(posBuf2, indBuf2, &posBuf1[i], tol);
            distMax = max(distMax, dist);
            if(dist > tol) {
                //cout << folder1 << " : file[" << k << "] : vertex[" << i << "] : dist=" << dist;
                match = false;
                break;
            }
        }
        cout << "dist=" << distMax << endl;
        if(!match) {
            break;
        }
        
        ++file1;
        ++file2;
    }
    return match;
}

int main(int argc, char **argv)
{
    if(argc < 3) {
        cout << "Usage: objscmp <FOLDER1> <FOLDER2> <OPTIONAL TOL>" << endl;
        return 0;
    }

    string folder1(argv[1]);
    string folder2(argv[2]);
    double tol = 1e-4;
    if(argc >= 4) {
        tol = atof(argv[3]);
    }

    bool pass = objscmp(folder1, folder2, tol);
    cout << (pass ? "PASS" : "FAIL") << endl;

    return 0;
}
